# @generated

# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# pylint: disable=protected-access,no-name-in-module

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: multi_pex_api.proto
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="multi_pex_api.proto",
    package="multi_pex_api",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=(
        b'\n\x13multi_pex_api.proto\x12\rmulti_pex_api"\x07\n\x05\x45mpty"8\n\x16\x43reatePexServerRequest\x12\x1e\n\x16\x63reate_pex_server_args\x18\x01'
        b' \x01(\t":\n\x14\x43reatePexServerReply\x12"\n\x1a\x63reate_pex_server_response\x18\x01'
        b' \x01(\t"4\n\x14GetPexServersRequest\x12\x1c\n\x14get_pex_servers_args\x18\x01'
        b' \x01(\t"6\n\x12GetPexServersReply\x12 \n\x18get_pex_servers_response\x18\x01'
        b' \x01(\t"<\n\x18ShutdownPexServerRequest\x12 \n\x18shutdown_pex_server_args\x18\x01'
        b' \x01(\t">\n\x16ShutdownPexServerReply\x12$\n\x1cshutdown_pex_server_response\x18\x01'
        b' \x01(\t"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01'
        b' \x01(\t"\x19\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01'
        b' \x01(\t2\xf0\x02\n\x0bMultiPexApi\x12_\n\x0f\x43reatePexServer\x12%.multi_pex_api.CreatePexServerRequest\x1a#.multi_pex_api.CreatePexServerReply"\x00\x12Y\n\rGetPexServers\x12#.multi_pex_api.GetPexServersRequest\x1a!.multi_pex_api.GetPexServersReply"\x00\x12\x65\n\x11ShutdownPexServer\x12\'.multi_pex_api.ShutdownPexServerRequest\x1a%.multi_pex_api.ShutdownPexServerReply"\x00\x12>\n\x04Ping\x12\x1a.multi_pex_api.PingRequest\x1a\x18.multi_pex_api.PingReply"\x00\x62\x06proto3'
    ),
)


_EMPTY = _descriptor.Descriptor(
    name="Empty",
    full_name="multi_pex_api.Empty",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=38,
    serialized_end=45,
)


_CREATEPEXSERVERREQUEST = _descriptor.Descriptor(
    name="CreatePexServerRequest",
    full_name="multi_pex_api.CreatePexServerRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="create_pex_server_args",
            full_name="multi_pex_api.CreatePexServerRequest.create_pex_server_args",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=47,
    serialized_end=103,
)


_CREATEPEXSERVERREPLY = _descriptor.Descriptor(
    name="CreatePexServerReply",
    full_name="multi_pex_api.CreatePexServerReply",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="create_pex_server_response",
            full_name="multi_pex_api.CreatePexServerReply.create_pex_server_response",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=105,
    serialized_end=163,
)


_GETPEXSERVERSREQUEST = _descriptor.Descriptor(
    name="GetPexServersRequest",
    full_name="multi_pex_api.GetPexServersRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="get_pex_servers_args",
            full_name="multi_pex_api.GetPexServersRequest.get_pex_servers_args",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=165,
    serialized_end=217,
)


_GETPEXSERVERSREPLY = _descriptor.Descriptor(
    name="GetPexServersReply",
    full_name="multi_pex_api.GetPexServersReply",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="get_pex_servers_response",
            full_name="multi_pex_api.GetPexServersReply.get_pex_servers_response",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=219,
    serialized_end=273,
)


_SHUTDOWNPEXSERVERREQUEST = _descriptor.Descriptor(
    name="ShutdownPexServerRequest",
    full_name="multi_pex_api.ShutdownPexServerRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="shutdown_pex_server_args",
            full_name="multi_pex_api.ShutdownPexServerRequest.shutdown_pex_server_args",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=275,
    serialized_end=335,
)


_SHUTDOWNPEXSERVERREPLY = _descriptor.Descriptor(
    name="ShutdownPexServerReply",
    full_name="multi_pex_api.ShutdownPexServerReply",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="shutdown_pex_server_response",
            full_name="multi_pex_api.ShutdownPexServerReply.shutdown_pex_server_response",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=337,
    serialized_end=399,
)


_PINGREQUEST = _descriptor.Descriptor(
    name="PingRequest",
    full_name="multi_pex_api.PingRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="echo",
            full_name="multi_pex_api.PingRequest.echo",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=401,
    serialized_end=428,
)


_PINGREPLY = _descriptor.Descriptor(
    name="PingReply",
    full_name="multi_pex_api.PingReply",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="echo",
            full_name="multi_pex_api.PingReply.echo",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=430,
    serialized_end=455,
)

DESCRIPTOR.message_types_by_name["Empty"] = _EMPTY
DESCRIPTOR.message_types_by_name["CreatePexServerRequest"] = _CREATEPEXSERVERREQUEST
DESCRIPTOR.message_types_by_name["CreatePexServerReply"] = _CREATEPEXSERVERREPLY
DESCRIPTOR.message_types_by_name["GetPexServersRequest"] = _GETPEXSERVERSREQUEST
DESCRIPTOR.message_types_by_name["GetPexServersReply"] = _GETPEXSERVERSREPLY
DESCRIPTOR.message_types_by_name["ShutdownPexServerRequest"] = _SHUTDOWNPEXSERVERREQUEST
DESCRIPTOR.message_types_by_name["ShutdownPexServerReply"] = _SHUTDOWNPEXSERVERREPLY
DESCRIPTOR.message_types_by_name["PingRequest"] = _PINGREQUEST
DESCRIPTOR.message_types_by_name["PingReply"] = _PINGREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Empty = _reflection.GeneratedProtocolMessageType(
    "Empty",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTY,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.Empty)
    },
)
_sym_db.RegisterMessage(Empty)

CreatePexServerRequest = _reflection.GeneratedProtocolMessageType(
    "CreatePexServerRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEPEXSERVERREQUEST,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.CreatePexServerRequest)
    },
)
_sym_db.RegisterMessage(CreatePexServerRequest)

CreatePexServerReply = _reflection.GeneratedProtocolMessageType(
    "CreatePexServerReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEPEXSERVERREPLY,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.CreatePexServerReply)
    },
)
_sym_db.RegisterMessage(CreatePexServerReply)

GetPexServersRequest = _reflection.GeneratedProtocolMessageType(
    "GetPexServersRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPEXSERVERSREQUEST,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.GetPexServersRequest)
    },
)
_sym_db.RegisterMessage(GetPexServersRequest)

GetPexServersReply = _reflection.GeneratedProtocolMessageType(
    "GetPexServersReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPEXSERVERSREPLY,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.GetPexServersReply)
    },
)
_sym_db.RegisterMessage(GetPexServersReply)

ShutdownPexServerRequest = _reflection.GeneratedProtocolMessageType(
    "ShutdownPexServerRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SHUTDOWNPEXSERVERREQUEST,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.ShutdownPexServerRequest)
    },
)
_sym_db.RegisterMessage(ShutdownPexServerRequest)

ShutdownPexServerReply = _reflection.GeneratedProtocolMessageType(
    "ShutdownPexServerReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _SHUTDOWNPEXSERVERREPLY,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.ShutdownPexServerReply)
    },
)
_sym_db.RegisterMessage(ShutdownPexServerReply)

PingRequest = _reflection.GeneratedProtocolMessageType(
    "PingRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _PINGREQUEST,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.PingRequest)
    },
)
_sym_db.RegisterMessage(PingRequest)

PingReply = _reflection.GeneratedProtocolMessageType(
    "PingReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _PINGREPLY,
        "__module__": "multi_pex_api_pb2"
        # @@protoc_insertion_point(class_scope:multi_pex_api.PingReply)
    },
)
_sym_db.RegisterMessage(PingReply)


_MULTIPEXAPI = _descriptor.ServiceDescriptor(
    name="MultiPexApi",
    full_name="multi_pex_api.MultiPexApi",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=458,
    serialized_end=826,
    methods=[
        _descriptor.MethodDescriptor(
            name="CreatePexServer",
            full_name="multi_pex_api.MultiPexApi.CreatePexServer",
            index=0,
            containing_service=None,
            input_type=_CREATEPEXSERVERREQUEST,
            output_type=_CREATEPEXSERVERREPLY,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetPexServers",
            full_name="multi_pex_api.MultiPexApi.GetPexServers",
            index=1,
            containing_service=None,
            input_type=_GETPEXSERVERSREQUEST,
            output_type=_GETPEXSERVERSREPLY,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="ShutdownPexServer",
            full_name="multi_pex_api.MultiPexApi.ShutdownPexServer",
            index=2,
            containing_service=None,
            input_type=_SHUTDOWNPEXSERVERREQUEST,
            output_type=_SHUTDOWNPEXSERVERREPLY,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="Ping",
            full_name="multi_pex_api.MultiPexApi.Ping",
            index=3,
            containing_service=None,
            input_type=_PINGREQUEST,
            output_type=_PINGREPLY,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_MULTIPEXAPI)

DESCRIPTOR.services_by_name["MultiPexApi"] = _MULTIPEXAPI

# @@protoc_insertion_point(module_scope)
